/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */  

#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <sys/socket.h>
#include <util/locker.h>
#include "atheossupport/MessageTransceiverThread.h"
#include "atheossupport/ThreadSafeObjectPool.h"
#include "dataio/TCPSocketDataIO.h"
#include "iogateway/PortableMessageIOGateway.h"
#include "util/NetworkUtilityFunctions.h"

static void ClearMessageFunc(PortableMessage * msg, void * userData);
void ClearMessageFunc(PortableMessage * msg, void * /*userData*/) {if (msg) {msg->what = 0; msg->MakeEmpty();}}

ThreadSafeObjectPool<PortableMessage> MessageTransceiverThread::_msgPool(100, ClearMessageFunc);
ThreadSafeObjectPool<PortableMessageRef::RefCountMem> MessageTransceiverThread::_msgRefPool;

MessageTransceiverThread ::
MessageTransceiverThread(int32 threadPri) : 
   _tcpThread(-1),
   _lock("mtt_lock"),
   _notifyMsg(PORTABLE_MESSAGES_RECEIVED), 
   _gateway(NULL),
   _datafd(-1),
   _acceptfd(-1),
   _notifyfd(-1),
   _waitfd(-1),
   _port(-1),
   _threadPri(threadPri),
   _isConnecting(false),
   _isServer(false),
   _sendOnConnect(false),
   _sendOnDisconnect(false),
   _sendOnEmpty(false), 
   _acceptMultiple(false),
   _messagesPendingSemaphore(-1),
   _messagesPendingBenaphore(0),
   _userGatewayFunc(NULL)
{
   // empty
}

MessageTransceiverThread ::
MessageTransceiverThread(const os::Messenger & replyTarget, int32 threadPri) :
   _tcpThread(-1),
   _replyTarget(replyTarget),
   _notifyMsg(PORTABLE_MESSAGES_RECEIVED),
   _lock("mtt_lock"),
   _gateway(NULL),
   _datafd(-1),
   _acceptfd(-1),
   _notifyfd(-1),
   _waitfd(-1),
   _port(-1),
   _threadPri(threadPri),
   _isConnecting(false), 
   _isServer(false),
   _sendOnConnect(false),
   _sendOnDisconnect(false),
   _sendOnEmpty(false),
   _acceptMultiple(false), 
   _messagesPendingSemaphore(-1), 
   _messagesPendingBenaphore(0),
   _userGatewayFunc(NULL)
{
   // empty
}

MessageTransceiverThread ::
~MessageTransceiverThread()
{
   // Kill the thread!
   thread_id thread = _tcpThread;
   if (thread >= 0)
   {
      _tcpThread = -1;
      NotifyThread();
      (void) wait_for_thread(thread);
   }
   delete _gateway;  // this closes the _datafd
   if (_acceptfd >= 0) closesocket(_acceptfd);
   if (_notifyfd >= 0) closesocket(_notifyfd);
   if (_waitfd  >= 0) closesocket(_waitfd);
   if (_messagesPendingSemaphore >= 0) delete_semaphore(_messagesPendingSemaphore);

   // For server mode, close any accepted connections that were never "picked up".
   for (int i=_incomingConnections.CountItems()-1; i>=0; i--) closesocket(_incomingConnections[i]);
}

status_t 
MessageTransceiverThread ::
StartConnectThread(const char * hostname, int port, const os::Message * optConnectMessage, const os::Message * optDisconnectMessage)
{
   if (CreateNotificationMechanism() != B_NO_ERROR) return B_ERROR;

   _connectTo = hostname;
   _port = port;

   if (optConnectMessage)
   {
      _onConnect = *optConnectMessage;
      _sendOnConnect = true;
   }
   if (optDisconnectMessage)
   {
      _onDisconnect = *optDisconnectMessage;
      _sendOnDisconnect = true;
   }

   return StartThread((void * const) ConnectThreadStub, "MessageTransceiver Connect");
}

status_t
MessageTransceiverThread ::
StartServerThread(int port, const os::Message & connectMessage, bool acceptMultiple)
{
   _isServer = true;
   status_t ret = StartAcceptThread(port, acceptMultiple, &connectMessage, NULL);
   if (ret != B_NO_ERROR) _isServer = false;  // back out of our commitment...
   return ret;
}

status_t 
MessageTransceiverThread ::
StartAcceptThread(int port, bool acceptMultiple, const os::Message * optConnectMessage, const os::Message * optDisconnectMessage )
{
   if (CreateNotificationMechanism() != B_NO_ERROR) return B_ERROR;

   _acceptMultiple = acceptMultiple;

   if (optConnectMessage)
   {
      _onConnect = *optConnectMessage;
      _sendOnConnect = true;
   }
   if (optDisconnectMessage)
   {
      _onDisconnect = *optDisconnectMessage;
      _sendOnDisconnect = true;
   }

   struct sockaddr_in saSocket;
   memset(&saSocket, 0, sizeof(saSocket));
   saSocket.sin_family      = AF_INET;
   saSocket.sin_addr.s_addr = htonl(INADDR_ANY);
   saSocket.sin_port        = htons(port);

   _acceptfd = socket(AF_INET, SOCK_STREAM, 0);
   if (_acceptfd >= 0)
   {
      const long trueValue = 1;
      (void) setsockopt(_acceptfd, SOL_SOCKET, SO_REUSEADDR, &trueValue, sizeof(long));
      if ((bind(_acceptfd, (struct sockaddr *) &saSocket, sizeof(saSocket)) == 0)&&(listen(_acceptfd, _acceptMultiple?20:1) == 0))
      {
         struct sockaddr_in waSocket;
         socklen_t len = sizeof(waSocket);
         if (getsockname(_acceptfd, (struct sockaddr *)&waSocket, &len) == 0)
         {
            _port = ntohs(waSocket.sin_port);
            return StartThread((void * const) TCPThreadStub, _isServer ? "MessageTransceiver Server"
                                                                       : "MessageTransceiver Accept");
         }
      }
      closesocket(_acceptfd);
      _acceptfd = -1;
   }                                                  
   return B_ERROR;
}

status_t 
MessageTransceiverThread ::
StartThread(AbstractMessageIOGateway * gateway, int datafd, const os::Message * optDisconnectMessage)
{
   if (CreateNotificationMechanism() != B_NO_ERROR) return B_ERROR;
   if (optDisconnectMessage)
   {
      _onDisconnect = *optDisconnectMessage;
      _sendOnDisconnect = true;
   }
   _gateway = gateway;
   _gateway->SetPools(GetMessagePool(), GetMessageRefPool());
   _datafd = datafd;
   return StartThread((void * const)TCPThreadStub, "MessageTransceiver Socket");
}

status_t
MessageTransceiverThread ::
StartThread(void * const func, const char * threadName)
{
   if (_tcpThread >= 0) return B_ERROR;  // only one thread at a time, pleez

   _messagesPendingSemaphore = create_semaphore("messages pending", 0, 0);
   if (_messagesPendingSemaphore >= 0)
   {
      _tcpThread = spawn_thread(threadName, func, _threadPri, 8192, this);
      if (_tcpThread >= 0) 
      {
         resume_thread(_tcpThread);
         return B_NO_ERROR;
      }   
      _tcpThread = -1;
      delete_semaphore(_messagesPendingSemaphore);
      _messagesPendingSemaphore = -1;
   }
   return B_ERROR;
}

status_t 
MessageTransceiverThread ::
GetNextIncomingMessage(PortableMessageRef & setRef)
{
   os::AutoLocker m(&_lock);
   return _gateway ? _gateway->GetNextIncomingMessage(setRef) : B_ERROR;
}

status_t
MessageTransceiverThread ::
GetNextIncomingConnection(int & returnfd)
{
   os::AutoLocker m(&_lock);
   if (_incomingConnections.CountItems() > 0)
   {
      _incomingConnections.RemoveHead(returnfd);
      return B_NO_ERROR;
   }
   else return B_ERROR;
}

void
MessageTransceiverThread ::
SetOutgoingQueueDrainedMessage(const os::Message * msg)
{
   os::AutoLocker m(&_lock);
   _sendOnEmpty = (msg != NULL);
   if (_sendOnEmpty) _onEmpty = *msg;
}

status_t 
MessageTransceiverThread ::
SendOutgoingMessage(const PortableMessageRef & msgRef)
{
   os::AutoLocker m(&_lock);
   if (_gateway)
   {
      bool messagesWerePending = _gateway->HasBytesToOutput();
      if (messagesWerePending == false) AcquireMessagesPendingLock();
      if (_gateway->AddOutgoingMessage(msgRef) != B_NO_ERROR) 
      {
         ReleaseMessagesPendingLock();
         return B_ERROR;
      }
      NotifyThread();
      return B_NO_ERROR;
   }
   return B_ERROR;
}

void 
MessageTransceiverThread ::
WaitForAllMessagesToBeSent()
{
   AcquireMessagesPendingLock();
   ReleaseMessagesPendingLock();
}

int 
MessageTransceiverThread ::
GetPort() const
{
   return _port;
}

// Creates our silly TCP-thread-wakeup mechanism.
// What we do is connect a TCP connection to ourself, and whenever
// we want to signal the TCP thread, we send a byte over this connection.
// This causes the TCP thread's select() to wake up, and then the TCP thread
// will do what we wanted.
status_t
MessageTransceiverThread ::
CreateNotificationMechanism()
{
   if (_waitfd >= 0) return B_NO_ERROR;  // already set up, eh?
   return CreateConnectedSocketPair(_waitfd, _notifyfd);
}

AbstractMessageIOGateway *
MessageTransceiverThread ::
CreateIOGateway(int socket) const
{
   AbstractMessageIOGateway * g = NULL;
   TCPSocketDataIO * io = newnothrow TCPSocketDataIO(socket, false);
   if (io)
   {
      g = _userGatewayFunc ? _userGatewayFunc(io) : newnothrow PortableMessageIOGateway(io);
      if (g) return g;
      io->ReleaseSocket();
      delete io;
   }
   WARN_OUT_OF_MEMORY;
   return g;
}

void
MessageTransceiverThread ::
ConnectThreadLoop()
{
   struct hostent * he = gethostbyname(_connectTo.Cstr());
   if (he)
   {
      int s = socket(AF_INET, SOCK_STREAM, 0);
      if (s >= 0)
      {
         struct sockaddr_in saAddr;
         memset(&saAddr, 0, sizeof(saAddr));
         saAddr.sin_family = AF_INET;
         saAddr.sin_port   = htons(_port);
         memcpy(&saAddr.sin_addr.s_addr, he->h_addr, he->h_length);

         if (SetSocketBlockingEnabled(s, false) == B_NO_ERROR)         
         {
            int ret = connect(s, (struct sockaddr *) &saAddr, sizeof(saAddr));
            bool inProgress = ((ret < 0)&&(errno == EINPROGRESS));
            if ((ret == 0)||(inProgress)) 
            {
               _gateway = CreateIOGateway(s);
               if (_gateway)
               {
                  _gateway->SetPools(GetMessagePool(), GetMessageRefPool());
                  _datafd = s;
                  _isConnecting = inProgress;
                  if ((_isConnecting == false)&&(_sendOnConnect)) _replyTarget.SendMessage(&_onConnect);
               }
            }
         }
         if (_datafd < 0) closesocket(s);
      }
   }

   // If we had a problem, let the user know the connect attempt failed
   if ((_datafd < 0)&&(_sendOnDisconnect)) _replyTarget.SendMessage(&_onDisconnect);

   TCPThreadLoop();
}

void 
MessageTransceiverThread ::
TCPThreadLoop()
{
   signal(SIGPIPE, SIG_IGN);  // avoid evil SIGPIPE signals from sending on a closed socket

   const thread_id me = get_thread_id(NULL);

   fd_set readSet;
   fd_set writeSet;
   while(me == _tcpThread)
   {
      FD_ZERO(&readSet);
      FD_ZERO(&writeSet);

      int maxfd = _waitfd;
      FD_SET(_waitfd, &readSet);
      
      if ((_gateway)&&(_datafd >= 0))
      {
         if (_datafd > maxfd) maxfd = _datafd;
         FD_SET(_datafd, &readSet);
         if ((_isConnecting)||(_gateway->HasBytesToOutput())) FD_SET(_datafd, &writeSet);
      }
      else
      {
         // If there is no current connection, and we have an accept socket,
         // then watch for incoming connections
         if (_acceptfd >= 0)
         {
            if (_acceptfd > maxfd) maxfd = _acceptfd;
            FD_SET(_acceptfd, &readSet);
         }
      }

      if (select(maxfd+1, &readSet, &writeSet, NULL, NULL) < 0) LogTime(MUSCLE_LOG_CRITICALERROR, "TCPThreadLoop: ERROR, select() failed!\n");
   
      if (FD_ISSET(_waitfd, &readSet))
      {
         char temp;
         (void) recv(_waitfd, &temp, sizeof(temp), 0L);
      }

      os::AutoLocker m(&_lock);
      bool hadMessagesReady = (_gateway)&&(_gateway->HasIncomingMessagesReady());
      bool sendDisconnectMessage = false;
      if ((_gateway)&&(_datafd >= 0))
      {
         bool readReady  = FD_ISSET(_datafd, &readSet);
         bool writeReady = FD_ISSET(_datafd, &writeSet);

         bool messagesWerePending = _gateway->HasBytesToOutput();
         bool readError  = ((readReady)&&(_gateway->DoInput() != B_NO_ERROR));
         bool writeError = ((writeReady)&&(_gateway->DoOutput() != B_NO_ERROR));
         if ((readError)||(writeError))
         {
            _datafd = -1;
            _gateway->GetDataIO()->Shutdown();  // close the connection to the client
            sendDisconnectMessage = _sendOnDisconnect;
         }
         else if ((_isConnecting)&&((readReady)||(writeReady)))
         {
            _isConnecting = false;
            {
               // This extra connect() is to tell select() that we got the message,
               // and he doesn't need to read-select us anymore about the connection
               // being established.  I hope this is documented behavior, and not
               // just a fortuitous side effect...
               struct sockaddr_in blah;
               memset(&blah, 0, sizeof(blah));
               connect(_datafd, (struct sockaddr *) &blah, sizeof(blah));
            }
            if (_sendOnConnect) _replyTarget.SendMessage(&_onConnect);
         }

         if ((messagesWerePending)&&((_gateway == NULL)||(_gateway->HasBytesToOutput() == false))) 
         {
            if (_sendOnEmpty) _replyTarget.SendMessage(&_onEmpty);
            ReleaseMessagesPendingLock();
         }
      }
      else 
      {
         if ((_acceptfd >= 0)&&(FD_ISSET(_acceptfd, &readSet)))
         {
            struct sockaddr_in junk;
            socklen_t nLen = sizeof(junk);
            int newSocket = accept(_acceptfd, (struct sockaddr *) &junk, &nLen);
            if (newSocket >= 0)
            {
               if (_isServer)
               {
                  if (_incomingConnections.AddTail(newSocket))
                  {
                     if (_sendOnConnect) _replyTarget.SendMessage(&_onConnect);
                  }
                  else closesocket(newSocket);
               }
               else
               {
                  delete _gateway;  // paranoia... but just in case there was one somehow
                  _gateway = CreateIOGateway(newSocket);
                  if (_gateway)
                  {
                     _gateway->SetPools(GetMessagePool(), GetMessageRefPool());
                     _datafd = newSocket;
                     if (_sendOnConnect) _replyTarget.SendMessage(&_onConnect);
                  }
                  else
                  {
                     WARN_OUT_OF_MEMORY;
                     closesocket(newSocket);
                  }
               }
            }
            if ((_isServer == false)&&(_gateway == NULL)) sendDisconnectMessage = _sendOnDisconnect;
            if (_acceptMultiple == false)
            {
               closesocket(_acceptfd);
               _acceptfd = -1;
            }
         }
      }
      if ((hadMessagesReady == false)&&((_gateway)&&(_gateway->HasIncomingMessagesReady()))) _replyTarget.SendMessage(&_notifyMsg);
      if (sendDisconnectMessage) _replyTarget.SendMessage(&_onDisconnect);
   }
}

void
MessageTransceiverThread ::
NotifyThread()
{
   if (_notifyfd >= 0) 
   {
      char temp = '!';
      (void) send(_notifyfd, &temp, sizeof(temp), 0L);
   }
}

void
MessageTransceiverThread ::
SetReplyTarget(const os::Messenger & target)
{
   os::AutoLocker m(&_lock);
   _replyTarget = target;
}

void
MessageTransceiverThread ::
SetGatewayFactoryFunc(GatewayFactoryFunc f)
{
   _userGatewayFunc = f;
}


void
MessageTransceiverThread ::
AcquireMessagesPendingLock()
{
   int32 previous = atomic_add(&_messagesPendingBenaphore, 1);
   if (previous > 0) lock_semaphore(_messagesPendingSemaphore);
}

void
MessageTransceiverThread ::
ReleaseMessagesPendingLock()
{
   int32 previous = atomic_add(&_messagesPendingBenaphore, -1);
   if (previous > 1) unlock_semaphore(_messagesPendingSemaphore);
}

