/* This file is Copyright 2001 Level Control Systems.  See the included LICENSE.txt file for details. */

#ifndef PORTABLE_MESSAGE_IO_GATEWAY
#define PORTABLE_MESSAGE_IO_GATEWAY

#include "dataio/PortableDataIO.h"
#include "message/PortableMessage.h"
#include "util/PortableRefCount.h"
#include "iogateway/AbstractMessageIOGateway.h"

/**
 * Encoding IDs.  Only PORTABLE_MESSAGE_ENCODING_DEFAULT is supported for now!
 */
enum {
   PORTABLE_MESSAGE_ENCODING_DEFAULT = 1164862256  // 'Enc0',  /**< just plain ol' flattened PortableMessage objects, with no special encoding */
};

/** Callback function type for flatten/unflatten notification callbacks */
typedef void (*MessageFlattenedCallback)(PortableMessageRef msgRef, void * userData);

/**
 * A "gateway" object that knows how to send/receive PortableMessages over a wire
 * (via a provided PortableDataIO object). 
 * May be subclassed to change the byte-level protocol, or used as is if the default protocol is desired.
 *
 * The default protocol format used by this class is:
 *   -# 4 bytes (uint32) indicating the flattened size of the message
 *   -# 4 bytes (uint32) indicating the encoding type (should always be PORTABLE_MESSAGE_ENCODING_DEFAULT for now)
 *   -# n bytes of flattened PortableMessage (where n is the value specified in 1)
 *   -# goto 1 ...
 */
class PortableMessageIOGateway : public AbstractMessageIOGateway
{
public:
   /** 
    *  Constructor.
    *  @param io The PortableDataIO object to read/write data through.  This PortableMessageIOGateway
    *            becomes the owner of (io) and will delete it later.
    *  @param encoding The byte-stream format the message should be encoded into.
    *                  The only supported format is PORTABLE_MESSAGE_ENCODING_DEFAULT, for now.
    */
   PortableMessageIOGateway(PortableDataIO * io, long encoding = PORTABLE_MESSAGE_ENCODING_DEFAULT);

   /**
    *  Destructor.
    *  Deletes the held PortableDataIO object.
    */
   virtual ~PortableMessageIOGateway();

   /**
    * Writes some of our outgoing messages to the wire.
    * Not guaranteed to write all outgoing messages (it will try not to block)
    * @return B_NO_ERROR usually, or B_ERROR if the connection has been broken
    *         or some other catastrophic condition has occurred.
    */
   virtual status_t DoOutput();

   /**
    * Reads some more bytes from the wire.  
    * If enough bytes have been read to assemble one or more new PortableMessages,
    * the new PortableMessages will be available after this method returns by
    * calling GetNextIncomingMessage().  
    * Tries not to block, but may (depending on implementation)
    * @return B_NO_ERROR usually, or B_ERROR if the connection has been broken
    *         or some other catastrophic condition has occurred.
    */
   virtual status_t DoInput();

   /**
    * Returns true iff there are any PortableMessages that have been
    * added (with AddOutgoingMessage()) but not already sent across the PortableDataIO.
    * Always returns false if the connection has been closed or hosed.
    */
   virtual bool HasBytesToOutput() const;

   /**
    * Lets you specify a function that will be called every time an outgoing
    * PortableMessage has been flattened by this gateway.
    * @param cb Callback function to call.
    * @param ud User data; set this to any value you like.
    */
   void SetMessageFlattenedCallback(MessageFlattenedCallback cb, void * ud) {_flattenedCallback = cb; _flattenedCallbackData = ud;}

   /**
    * Lets you specify a function that will be called every time an incoming
    * PortableMessage has been unflattened by this gateway.
    * @param cb Callback function to call.
    * @param ud User data; set this to any value you like.
    */
   void SetMessageUnflattenedCallback(MessageFlattenedCallback cb, void * ud) {_unflattenedCallback = cb; _unflattenedCallbackData = ud;}

protected:
   /**
    * Should return the total number of bytes to use for the body buffer for flattening the given message.
    * @param msg The PortableMessage in question.
    * @return Default implementation returns msg.FlattenedSize().
    */
   virtual size_t FlattenedMessageBodySize(const PortableMessage & msg) const;

   /**
    * Flattens (msg) into (buffer) and (header).
    * @param msg The PortableMessage to flatten into a byte array.
    * @param header A byte array that is HeaderSize() bytes long.  FlattenMessage() must write the
    *               proper header bytes into this array.
    * @param buffer A byte array that is FlattenedMessageBodySize() bytes long.  FlattenMessage()
    *               must write the proper message body bytes into this array.
    * @return B_NO_ERROR if the flattening succeeded and the produced bytes should be sent, 
    *         B_ERROR if flattening failed and/or the produced bytes should not be sent.
    * The default implementation always returns B_NO_ERROR.
    */
   virtual status_t FlattenMessage(const PortableMessage & msg, uint8 * header, uint8 * buffer) const;

   /**
    * Unflattens (msg) from (buffer) and (header).
    * @param setMsg An empty PortableMessage.  On success, the restored PortableMessage should be written into this object.
    * @param header The buffer of message header bytes.  It is HeaderSize() bytes long.
    * @param buf The buffer of message body bytes.  It is (bufSize) bytes long.
    * @param bufSize The number of bytes pointed to by (buf)
    * @return B_NO_ERROR on success, B_ERROR if the PortableMessage couldn't be restored.
    */
   virtual status_t UnflattenMessage(PortableMessage & setMsg, const uint8 * header, const uint8 * buf, size_t bufSize) const;
 
   /**
    * Returns the size of the pre-flattened-message header section, in bytes.
    * The default encoding has an 8-byte header (4 bytes for encoding ID, 4 bytes for message size)
    */
   virtual size_t HeaderSize() const;

   /**
    * Must Extract and returns the buffer body size from the given header.
    * Note that the returned size should NOT include the header bytes!
    * @param header Points to the header of the message.  The header is HeaderSize() bytes long.
    * @param setSize On success, the number of bytes to expect in the body of the message should be written here.
    * @return B_NO_ERROR if successful, B_ERROR if the header format was incorrect or for some
    *         other reason the message body size could not be determined.
    */
   virtual status_t GetBodySize(const uint8 * header, size_t & setSize) const;

   /**
    * This method is called after GetBodySize() returns an error code.  (i.e.
    * when the header bytes we read in were not valid)
    * It should return B_NO_ERROR if the gateway should try to recover from
    * the invalid header, or B_ERROR if it should just give up and close the connection.
    * If it returns B_NO_ERROR, it may set the (setFirstValidByte) parameter to the
    * index of the first byte in the header buffer to start reading the next header
    * from, or leave (setFirstValidByte) unchanged to throw out all the bytes in the 
    * current header buffer and start afresh.
    * Default implementation just returns B_ERROR.
    * @param headerBuf the header bytes that were read in.
    * @param setFirstValidByte the length of the header.  May be set as described above.
    * @return B_NO_ERROR if recovery was successful, else B_ERROR.  (see above)
    */
   virtual status_t RecoverFromInvalidHeader(const uint8 * headerBuf, size_t & setFirstValidByte) const;
   
   /**
    * This method is called after UnflattenMessage() returns an error code (i.e.
    * when the body bytes we read in were not valid).
    * It should return B_NO_ERROR if the gateway should try to recover from the
    * invalid message, or B_ERROR if it should just give up and close the connection.
    * If B_NO_ERROR is returned, the the gateway will begin trying to read the
    * next header from the bytes immediately after these.
    * @param messageBuf Pointer to the message bytes.
    * @param messageBufSize Number of bytes pointed to by (messageBuf)
    * @return B_NO_ERROR to continue parsing, B_ERROR to give up (see above).
    */
   virtual status_t RecoverFromInvalidMessage(const uint8 * messageBuf, size_t messageBufSize) const;

private:
   uint8 * _sendBuffer;
   size_t  _sendBufferSize;
   int32   _sendBufferOffset;
   size_t  _currentSendMessageSize;  // includes header bytes

   uint8 * _recvBuffer;
   size_t  _recvBufferSize;
   int32   _recvBufferOffset;
   size_t  _currentRecvMessageSize;  // includes header bytes

   int     _encoding;
  
   MessageFlattenedCallback _flattenedCallback;
   void * _flattenedCallbackData;

   MessageFlattenedCallback _unflattenedCallback;
   void * _unflattenedCallbackData;
};

#endif
