/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */

#include <stdio.h>

#include "util/NetworkUtilityFunctions.h"
#include "support/PortableMacros.h"
#include "support/PortableStatusConstants.h"

#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netdb.h>
#include <netinet/in.h>

int Connect(const char * hostName, int port, const char * debugTitle, bool errorsOnly)
{
   if ((debugTitle)&&(errorsOnly == false))
   {
      LogTime(MUSCLE_LOG_INFO, "%s: Connecting to [%s:%i]: ", debugTitle, hostName, port);
      fflush(stdout);
   }
   struct hostent * he = gethostbyname(hostName);
   if (he)
   {
      int s = socket(AF_INET, SOCK_STREAM, 0);
      if (s >= 0)
      {
         struct sockaddr_in saAddr;
         memset(&saAddr, 0, sizeof(saAddr));
         saAddr.sin_family = AF_INET;
         saAddr.sin_port   = htons(port);
         memcpy(&saAddr.sin_addr.s_addr, he->h_addr, he->h_length);

         if (connect(s, (struct sockaddr *) &saAddr, sizeof(saAddr)) >= 0)
         {
            if ((debugTitle)&&(errorsOnly == false)) Log(MUSCLE_LOG_INFO, "Connected!\n");
            return s;
         }
         else if (debugTitle)
         {
            if (errorsOnly) Log(MUSCLE_LOG_INFO, "%s: connect() to [%s:%i] failed!\n", debugTitle, hostName, port);
                       else Log(MUSCLE_LOG_INFO, "Connection failed!\n");
         }
         closesocket(s);
      }
      else if (debugTitle)
      {
         if (errorsOnly) Log(MUSCLE_LOG_INFO, "%s: socket() failed!\n", debugTitle);
                    else Log(MUSCLE_LOG_INFO, "socket() failed!\n");
      }
   }
   else if (debugTitle)
   {
      if (errorsOnly) Log(MUSCLE_LOG_INFO,"%s: Couldn't lookup hostname [%s]\n", debugTitle, hostName);
                 else Log(MUSCLE_LOG_INFO,"hostname lookup failed!\n");
   }
   return -1;
}

int CreateAcceptSocket(int /*listenPort*/)
{
   return -1;  // TODO: implement this
}

void Inet_NtoA(uint32 addr, char * buf)
{
   sprintf(buf, "%li.%li.%li.%li", (addr>>24)&0xFF, (addr>>16)&0xFF, (addr>>8)&0xFF, (addr>>0)&0xFF);
}

status_t GetPeerInfo(int socket, uint32 & setIPAddress, uint16 & setPort)
{
   struct sockaddr_in saTempAdd;
   net_length_t length = sizeof(sockaddr_in);
   if (getpeername(socket, (struct sockaddr *)&saTempAdd, &length) == 0)
   {
      setIPAddress = ntohl(saTempAdd.sin_addr.s_addr);
      setPort = ntohs(saTempAdd.sin_port);
      return B_NO_ERROR;
   }
   return B_ERROR;
}


/* Source code stolen from UNIX Network Programming, Volume 1
 * Comments from the Unix FAQ
 */
status_t
BecomeDaemonProcess(const char * optNewDir, const char * optOutputTo, bool createIfNecessary)
{
   // Here are the steps to become a daemon:
   // 1. fork() so the parent can exit, this returns control to the command line or shell invoking
   //    your program. This step is required so that the new process is guaranteed not to be a process
   //    group leader. The next step, setsid(), fails if you're a process group leader.
   pid_t pid = fork();
   if (pid < 0) return B_ERROR;
   if (pid > 0) exit(0);

   // 2. setsid() to become a process group and session group leader. Since a controlling terminal is
   //    associated with a session, and this new session has not yet acquired a controlling terminal
   //    our process now has no controlling terminal, which is a Good Thing for daemons.
   setsid();

   // 3. fork() again so the parent, (the session group leader), can exit. This means that we, as a
   //    non-session group leader, can never regain a controlling terminal.
   signal(SIGHUP, SIG_IGN);
   pid = fork();
   if (pid < 0) return B_ERROR;
   if (pid > 0) exit(0);

   // 4. chdir("/") can ensure that our process doesn't keep any directory in use. Failure to do this
   //    could make it so that an administrator couldn't unmount a filesystem, because it was our
   //    current directory. [Equivalently, we could change to any directory containing files important
   //    to the daemon's operation.]
   if (optNewDir) chdir(optNewDir);

   // 5. umask(0) so that we have complete control over the permissions of anything we write.
   //    We don't know what umask we may have inherited. [This step is optional]
   umask(0);

   // 6. close() fds 0, 1, and 2. This releases the standard in, out, and error we inherited from our parent
   //    process. We have no way of knowing where these fds might have been redirected to. Note that many
   //    daemons use sysconf() to determine the limit _SC_OPEN_MAX. _SC_OPEN_MAX tells you the maximun open
   //    files/process. Then in a loop, the daemon can close all possible file descriptors. You have to
   //    decide if you need to do this or not. If you think that there might be file-descriptors open you should
   //    close them, since there's a limit on number of concurrent file descriptors.
   // 7. Establish new open descriptors for stdin, stdout and stderr. Even if you don't plan to use them,
   //    it is still a good idea to have them open. The precise handling of these is a matter of taste;
   //    if you have a logfile, for example, you might wish to open it as stdout or stderr, and open `/dev/null'
   //    as stdin; alternatively, you could open `/dev/console' as stderr and/or stdout, and `/dev/null' as stdin,
   //    or any other combination that makes sense for your particular daemon.
   int nullfd = open("/dev/null", O_RDWR);
   if (nullfd >= 0) dup2(nullfd, STDIN_FILENO);

   int outfd = -1;
   if (optOutputTo) 
   {
      outfd = open(optOutputTo, O_WRONLY | (createIfNecessary ? O_CREAT : 0));
      if (outfd < 0) LogTime(MUSCLE_LOG_ERROR, "BecomeDaemonProcess():  Couldn't open %s to redirect stdout, stderr\n", optOutputTo);
   }
   if (outfd >= 0) dup2(outfd, STDOUT_FILENO);
   if (outfd >= 0) dup2(outfd, STDERR_FILENO);

   return B_NO_ERROR;
}


/* See the header file for description of what this does */
status_t CreateConnectedSocketPair(int & socket1, int & socket2, bool blocking)
{
   if ((socket1 = socket(AF_INET, SOCK_STREAM, 0)) >= 0)
   {
      long trueValue = 1;
      if (setsockopt(socket1, SOL_SOCKET, SO_REUSEADDR, &trueValue, sizeof(long)) == 0)
      {
         struct sockaddr_in saSocket;
         memset(&saSocket, 0, sizeof(saSocket));
         saSocket.sin_family = AF_INET;
         if ((bind(socket1, (struct sockaddr *) &saSocket, sizeof(saSocket)) == 0)&&(listen(socket1, 20) == 0))
         {
            // Find out what port we're listening on
            struct sockaddr_in waSocket;
            net_length_t len = sizeof(waSocket);
     
            if (getsockname(socket1, (struct sockaddr *)&waSocket, &len) == 0)
            {
               if ((socket2 = socket(AF_INET, SOCK_STREAM, 0)) >= 0)
               {
                  struct sockaddr_in caSocket;
                  memset(&caSocket, 0, sizeof(caSocket));
                  caSocket.sin_family = AF_INET;
                  caSocket.sin_port   = waSocket.sin_port;
                  caSocket.sin_addr.s_addr = htonl((127<<24)|1);  // 127.0.0.1
                  if (connect(socket2, (struct sockaddr *) &caSocket, sizeof(caSocket)) == 0) 
                  {
                     struct sockaddr_in aaTempAd;
                     net_length_t alen = sizeof(aaTempAd);
                     int newfd = accept(socket1, (struct sockaddr *)&aaTempAd, &alen);
                     if (newfd >= 0)
                     {
                        closesocket(socket1);
                        socket1 = newfd;
                        if ((SetSocketBlockingEnabled(socket1, blocking) == B_NO_ERROR)&&
                            (SetSocketBlockingEnabled(socket2, blocking) == B_NO_ERROR)) return B_NO_ERROR;
                     }
                  }
               }
               closesocket(socket2);
               socket2 = -1;
            }
         }
      }
      closesocket(socket1);
      socket1 = -1;
   }
   return B_ERROR;
}

status_t
SetSocketBlockingEnabled(int socket, bool blocking)
{
#ifdef BEOS_OLD_NETSERVER
      long b = blocking ? 0 : 1; 
      return (setsockopt(socket, SOL_SOCKET, SO_NONBLOCK, &b, sizeof(b)) == 0) ? B_NO_ERROR : B_ERROR;
#else
      int flags = fcntl(socket, F_GETFL, 0);
      if (flags < 0) return B_ERROR;
      flags = blocking ? (flags&~O_NONBLOCK) : (flags|O_NONBLOCK);
      return (fcntl(socket, F_SETFL, flags) == 0) ? B_NO_ERROR : B_ERROR;
#endif
}
