/* This file is Copyright 2001 Level Control Systems.  See the included LICENSE.txt file for details. */

/* Note:  This file was originally written by Michael Olivero, but has been modified
 *        extensively.  Nonetheless, the original license is listed below.  --jaf
 */

/*
   This library was downloaded from: http://www.mike95.com

   This library is copyright.  It may freely be used for personal purposes 
   if the restriction listed below is adhered to.
       Author: Michael Olivero
       Email:  mike95@mike95.com

   //===============================
   //Start of Restriction Definition
   //===============================
   Anyone can have full use of the library provided they keep this complete comment
   with the source.  Also I would like to ask if any changes are made to the
   code for efficiency reasons, please let me know so I may look into your change and
   likewise incorporate it into library.  If a suggestion makes it into the library,
   your credits will be added to this information.

   Authors of Computer related books are welcome to include this source code as part
   of their publishing, provided credit the Author and make note of where the source
   code was obtained from: http://www.mike95.com
   //=============================
   //End of Restriction Definition
   //=============================

   Description:
   Visit http://www.mike95.com/c_plusplus/classes/JHashtable/

   This class is a based on the Java Hashtable class and as such contains all the public
   member functions of its Java equivalent.  Unlike Java, typecasts are not necessary
   since C++ allows template instatiation of types at compile time.

   Note:  Since java has hashCode() as a member of the base Object class, all 
   Java classes are inherently hashable.  Since the template parameter types do
   not necessarily have to have a built in hashing function, the user of the class
   must specify a hash function by calling setHashFunction() passing a pointer 
   to the hash function.

   The has function must be declared as the following:

       uint32 function(const KeyType&);

   Where:
   function = any name you choose to use for the function name
   KeyType  = the type used for the key in the construction of the PortableHashtable object.

   Example:
       uint32 myHash( const int&)
       {
           //your hashing code here for a key of type int.
       }
*/

#ifndef _PORTABLE_HASHTABLE_H
#define _PORTABLE_HASHTABLE_H

#include "support/PortableTypeDefs.h"
#include "support/PortableMacros.h"
#include "support/PortableStatusConstants.h"

template <class KeyType, class ValueType> class PortableHashtable;  // forward declaration

/**
 * This class is an iterator object, used for iterating over the set
 * of keys or values in a PortableHashtable. 
 *
 * Given a PortableHashtable object, you can obtain one or more of these
 * iterator objects by calling the PortableHashtable's GetIterator() method.
 *
 * This iterator actually contains separate state for two iterations:
 * one for iterating over the values in the PortableHashtable, and one for 
 * iterating over the keys.  These two iterations can be done independently
 * of each other.
 *
 * Note that it is illegal to modify the hashtable during an iteration 
 * traversal (an assertion failure will occur if you try it, and your 
 * program will exit)
 */
template <class KeyType, class ValueType> class PortableHashtableIterator
{
public:
   /**
    * Default constructor.  It's here only so that you can include PortableHashtableIterators
    * as member variables, in arrays, etc.  PortableHashtableIterators created with this
    * constructor are "empty", so they won't be useful until you set them equal to a 
    * PortableHashtableIterator that was returned by PortableHashtable::GetIterator().
    */
   PortableHashtableIterator();

   /** Copy Constructor.  */
   PortableHashtableIterator(const PortableHashtableIterator<KeyType, ValueType> & rhs);

   /** Destructor */
   ~PortableHashtableIterator();

   /** Returns true iff there are more keys left in the key traversal.  */
   bool HasMoreKeys() const;

   /**
    * Gets the next key in the key traversal.
    * @param setNextKey On success, the next key is copied into this object.
    * @return True on success, false if there are no more keys left.
    */
   bool GetNextKey(KeyType & setNextKey);

   /**
    * Gets a pointer to the next key in the key traversal.
    * Note that the returned pointer is only guaranteed valid as long as the Hashtable remains unchanged.
    * @return A pointer to the next key in the key traversal, or NULL if there are no more keys left.
    */
   KeyType * GetNextKey();

   /**
    * Peek at the next key without modifying the state of the traversal.
    * @param setKey On success, the next key is copied into this object.
    * @return True on success, or false if there are no more keys left.
    */
   bool PeekNextKey(KeyType & setKey) const;

   /**
    * Peek at the next key without modifying the state of the traversal.
    * Note that the returned pointer is only guaranteed valid as long as the Hashtable remains unchanged.
    * @return a pointer to the next key in the key traversal, or NULL if there are no keys left.
    */
   KeyType * PeekNextKey() const;

   /** Returns true iff there are more values left in the value traversal.  */
   bool HasMoreValues() const;

   /**
    * Get the next value in the values traversal.
    * @param setNextValue On success, the next value in the traversal is copied into this object.
    * @return True on success, false if there are no more values left in the value traversal.
    */
   bool GetNextValue(ValueType & setNextValue);

   /**
    * Get the next value in the values traversal.
    * Note that the returned pointer is only guaranteed valid as long as the Hashtable remains unchanged.
    * @return a pointer to the next value in the value traversal, or NULL if there are no values left.
    */
   ValueType * GetNextValue();

   /**
    * Peek at the next value without modifying the state of the traversal.
    * @param setValue On success, the next value is copied into this object.
    * @return True on success, or false if there are no more values left.
    */
   bool PeekNextValue(ValueType & setValue) const;

   /**
    * Peek at the next value without modifying the state of the traversal.
    * Note that the returned pointer is only guaranteed valid as long as the Hashtable remains unchanged.
    * @return a pointer to the next value in the value traversal, or NULL if there are no values left.
    */
   ValueType * PeekNextValue() const;

   /*
    * (note that the default copy constructor and '=' operator are okay for our purposes,
    *  as we don't "own" any of the data that we point to)
    */

private:
   friend class PortableHashtable<KeyType, ValueType>;

   PortableHashtableIterator(const PortableHashtable<KeyType, ValueType> & owner, uint32 stateCode);

   void * _nextKeyCookie;
   void * _nextValueCookie;

   uint32 _stateCode;  // used to detect modifications-during-iterator-traversals and trigger assertion failures

   const PortableHashtable<KeyType, ValueType> * _owner;
};


/**
 *  This is a handy templated Hashtable class, rather similar to Java's java.util.Hashtable.
 */
template <class KeyType, class ValueType> class PortableHashtable
{
public:
   /** Constructor.
    *  @param func Specified a hashing function that returns an integer hashcode for a given key object.
    *  @param initialCapacity Specifies the number of table slots to initially pre-allocate.   Defaults to 13.
    *  @param loadFactor Specifies how densely populated the table must be before it is re-hashed.  Defaults to 0.5 (i.e. 50%)
    */
   PortableHashtable(uint32(*func)(const KeyType& key), uint32 initialCapacity = 13, float loadFactor = 0.5f);

   /** Copy Constructor.  */
   PortableHashtable(const PortableHashtable<KeyType,ValueType> & rhs);

   /** Assignment operator. */
   PortableHashtable& operator=(const PortableHashtable<KeyType,ValueType> & rhs);

   /** Destructor.  */
   ~PortableHashtable();

   /** Returns the number of items in the table. */
   uint32 Size() const {return _count;}

   /** Returns true iff the table is empty. */
   bool IsEmpty() const {return _count == 0;}

   /** Returns true iff the table contains a mapping with the given value.  (O(n) search time) */
   bool Contains(const ValueType& value) const;

   /** Returns true iff the table contains a mapping with the given key.  (O(1) search time) */
   bool ContainsKey( const KeyType& key) const;

   /** Retrieve the associated value from the table for a given key.
    *  @param key The key to use to look up a value.
    *  @param setValue On success, the associated value is copied into this object.
    *  @return True on success, false if their was no value found for the given key.
    */
   bool Get(const KeyType& key, ValueType& setValue) const; 

   /** Retrieve a pointer to the associated value object for the given key.
    *  @param key The key to use to look up a value.
    *  @return A pointer to the internally held value object for the given key,
    *          or NULL of no object was found.  Note that this object is only
    *          guaranteed to remain valid as long as the Hashtable remains unchanged.
    */
   ValueType * Get(const KeyType & key) const;

   /** Returns an iterator object that can be used to examine the items in the hash table. */
   PortableHashtableIterator<KeyType,ValueType> GetIterator() const 
     {return PortableHashtableIterator<KeyType,ValueType>(*this, _stateCode);}

   /** Places the given (key, value) mapping into the table.  Any previous entry with
    *   a key of (key) will be replaced.  
    *  @param key The key that the new value is to be associated with.
    *  @param value The value to associate with the new key.
    *  @param setPreviousValue If there was a previously existing value associated with
    *                          (key), it will be copied into this object.
    *  @param optSetReplaced If set non-NULL, this boolean will be set to true if (setPreviousValue)
    *                        was written into, false otherwise.
    *  @return True If the operation succeeded, false if it failed (out of memory?)
    *               NOTE:  THIS BEHAVIOUR HAS CHANGED FROM THAT OF v1.64 AND BELOW!!!
    */
   bool Put(const KeyType& key, const ValueType& value, ValueType & setPreviousValue, bool * optSetReplaced = false);

   /** Places the given (key, value) mapping into the table.  Any previous entry with
    *  a key of (key) will be replaced.
    *  @param key The key that the new value is to be associated with.
    *  @param value The value to associate with the new key.
    *  @return True If the operation succeeded, false if it failed (out of memory?)
    *               NOTE:  THIS BEHAVIOUR HAS CHANGED FROM THAT OF v1.64 AND BELOW!!!
    */
   bool Put(const KeyType& key, const ValueType& value);

   /** Removes a mapping from the table.
    *  @param key The key of the key-value mapping to remove.
    *  @return True if a key was found and the mapping removed, or false if the key wasn't found.
    */
   bool Remove(const KeyType& key);

   /** Removes the mapping with the given (key) and places its value into (setRemovedValue).
    *  @param key The key of the key-value mapping to remove.
    *  @param setRemovedValue On success, the removed value is copied into this object.
    *  @return True if a key was found and the mapping removed, or false if the key wasn't found.
    */
   bool Remove(const KeyType& key, ValueType & setRemovedValue);

   /** Removes all mappings from the hash table. */
   void Clear();

   /** You can call this to change the hash function that is used to generate integer
    *  hashing values from the key items.
    *  @param func The new hashing function to use for this table. 
    */
   void SetHashFunction(uint32(*func)(const KeyType& key)) {_userHashFunc = func;}

   /** You can call this to set a function to use to compare keys.  By default, the 
    *  Key class's '==' operator is used, but for some key types (e.g. const char *'s) '==' is not
    *  useful, so you can supply your own here.
    *  @param func The new comparison function to use for this table.
    */
   void SetCompareFunction(bool(*func)(const KeyType& key1, const KeyType& key2)) {_userCompareFunc = func;}

private:
   friend class PortableHashtableIterator<KeyType, ValueType>;

   class PortableHashtableEntry
   {
   public:
      PortableHashtableEntry() : _next(NULL), _iterPrev(NULL), _iterNext(NULL), _valid(false)
      {
         // empty
      }
 
      PortableHashtableEntry(const PortableHashtableEntry & rhs) : _iterPrev(NULL), _iterNext(NULL)
      {
         *this = rhs;
      }

      ~PortableHashtableEntry()
      {
         // empty
      }

      PortableHashtableEntry & operator=(const PortableHashtableEntry & rhs)
      {
         _hash  = rhs._hash;
         _key   = rhs._key;
         _value = rhs._value;
         _next  = rhs._next;
         _valid = rhs._valid;
         // DO NOT copy _iterPrev and _iterNext from rhs!  They must remain the same as before.
         return * this;
      }
 
      // Returns the entry to its just-created state
      void Invalidate(const KeyType & defaultKey, const ValueType & defaultValue)
      {
         // Gotta reset, so that something like a RefCount object
         // as a KeyType or ValueType, doesn't sit around without
         // releasing its held object/memory for a loooonnng time...
         _key   = defaultKey;
         _value = defaultValue;
         _valid = false;
         _next = _iterPrev = _iterNext = NULL;
         // hash is undefined, so no point in resetting it here
      }

      uint32 _hash;                   // precalculated for efficiency
      KeyType _key;                   // used for '==' checking
      ValueType _value;               // payload
      PortableHashtableEntry* _next;  // for making linked lists in our bucket
      PortableHashtableEntry* _iterPrev;  // for table iteration
      PortableHashtableEntry* _iterNext;  // for table iteration
      bool _valid;                    // only used for entries in our _table array, (which are by value and so can't be NULL)
   };

   // Auxilliary methods
   bool PutAux(const KeyType& key, const ValueType& value, ValueType * optSetPreviousValue, bool * optReplacedFlag);
   bool RemoveAux(const KeyType& key, ValueType * setRemovedValue);

   // Linked list maintainence
   void AddIterationEntry(PortableHashtableEntry * e);
   void RemoveIterationEntry(PortableHashtableEntry * e);
   void RecalculateIterationEntries();

   // PortableHashtableIterator's private API
   KeyType * GetKeyFromCookie(void * c, uint32 sc) const 
   {
      CheckStateCode(sc);
      return c ? &(((PortableHashtableEntry *)c)->_key) : NULL;
   }

   ValueType * GetValueFromCookie(void * c, uint32 sc) const  
   {
      CheckStateCode(sc);
      return c ? &(((PortableHashtableEntry *)c)->_value) : NULL;
   }

   void IterateCookie(void ** c) const 
   {
      PortableHashtableEntry * entry = *((PortableHashtableEntry **)c);
      *c = entry ? entry->_iterNext : NULL;
   }

   void CheckStateCode(uint32 stateCode) const
   {
      if (stateCode != _stateCode)
      {
         char temp[100];
         sprintf(temp, "PortableHashtable modified during traversal (diff=%lu)", _stateCode-stateCode);
         MASSERT(false, temp);
      }
   }

   bool KeysAreEqual(const KeyType & key1, const KeyType & key2) const;
   inline uint32 NextPrime(uint32 start);
   bool GrowTable();

   uint32 _count;       //the size of the elements in the hashtable
   uint32 _tableSize;   //the size of the table.
   uint32 _threshold;
   float _loadFactor;
   PortableHashtableEntry * _table;
   PortableHashtableEntry * _iterHead;    // start of linked list to iterate through
   KeyType _defaultKey;      // used for invalidating PortableHashtableEntry's
   ValueType _defaultValue;  // used for invalidating PortableHashtableEntry's
   uint32 (*_userHashFunc)(const KeyType& key);
   bool (*_userCompareFunc)(const KeyType& key1, const KeyType& key2);
   uint32 _stateCode;  // used to detect modifications-during-iterator-traversals
};

//===============================================================
// Implementation of PortableHashtable
// Necessary location for appropriate template instantiation.
//===============================================================
template <class KeyType, class ValueType>
PortableHashtable<KeyType,ValueType>::PortableHashtable(uint32 (*func)(const KeyType & key), uint32 initialCapacity, float loadFactor)
   : _count(0), _tableSize(initialCapacity), _threshold((uint32)(initialCapacity*loadFactor)), _loadFactor(loadFactor), _table(NULL), _iterHead(NULL), _userHashFunc(func), _userCompareFunc(NULL), _stateCode(0)
{
   // empty
}

template <class KeyType, class ValueType>
PortableHashtable<KeyType,ValueType>::
PortableHashtable(const PortableHashtable<KeyType,ValueType> & rhs)
   : _count(0), _tableSize(rhs._tableSize), _threshold(rhs._threshold), _loadFactor(rhs._loadFactor), _table(NULL), _iterHead(NULL), _userHashFunc(rhs._userHashFunc), _userCompareFunc(rhs._userCompareFunc), _stateCode(0)
{
   *this = rhs;
}

template <class KeyType, class ValueType>
PortableHashtable<KeyType, ValueType> &
PortableHashtable<KeyType, ValueType> ::
operator=(const PortableHashtable<KeyType, ValueType> & rhs)
{
   if (this != &rhs)
   {
      Clear();
      PortableHashtableIterator<KeyType, ValueType> iter = rhs.GetIterator();
      KeyType * nextKey;
      while((nextKey = iter.GetNextKey()) != NULL) (void) Put(*nextKey, *iter.GetNextValue());  // no good way to handle out-of-mem here?  
   }
   return *this;
}

template <class KeyType, class ValueType>
PortableHashtable<KeyType,ValueType>::~PortableHashtable()
{
   Clear();
   delete [] _table;
}

template <class KeyType, class ValueType>
bool 
PortableHashtable<KeyType,ValueType>::KeysAreEqual(const KeyType & key1, const KeyType & key2) const
{
   return (_userCompareFunc) ? _userCompareFunc(key1, key2) : (key1 == key2);
}

template <class KeyType, class ValueType>
bool
PortableHashtable<KeyType,ValueType>::Contains(const ValueType& value) const
{
   PortableHashtableIterator<KeyType, ValueType> iter = GetIterator();
   ValueType * v;
   while((v = iter.GetNextValue()) != NULL) if (*v == value) return true;
   return false;
}

template <class KeyType, class ValueType>
bool
PortableHashtable<KeyType,ValueType>::ContainsKey(const KeyType& key) const
{
   return (Get(key) != NULL);
}

template <class KeyType, class ValueType>
bool
PortableHashtable<KeyType,ValueType>::Get(const KeyType& key, ValueType & setValue) const
{
   ValueType * ptr = Get(key);
   if (ptr)
   {
      setValue = *ptr;
      return true;
   }
   else return false;
}

template <class KeyType, class ValueType>
ValueType *
PortableHashtable<KeyType,ValueType>::Get(const KeyType& key) const
{
   if (_table)
   {
      uint32 hash = (*_userHashFunc)(key);
      PortableHashtableEntry * e = &_table[hash % _tableSize];
      if (e->_valid)
      {
         while(e)
         {
            if ((e->_hash == hash)&&(KeysAreEqual(e->_key, key))) return &e->_value;
            e = e->_next; 
         }
      }
   }
   return NULL;
}

template <class KeyType, class ValueType>
bool
PortableHashtable<KeyType,ValueType>::GrowTable()
{
   _stateCode++;

   if (_table)
   {
      uint32 newTableSize = NextPrime(2 * _tableSize);
      PortableHashtableEntry * newTable = newnothrow PortableHashtableEntry[newTableSize];
      if (newTable == NULL) {WARN_OUT_OF_MEMORY; return false;}

      // Dump the iteration list (we are going to munge all the _iterNext fields for scratch space anyway)
      _iterHead = NULL;
 
      // First, build up two singly-linked lists:  one is a list of entries embedded in our _table 
      // array, and the other is a list of entries that were dynamically allocated (in our buckets).
      PortableHashtableEntry * staticHead  = NULL;   // demand-allocated
      PortableHashtableEntry * dynamicHead = NULL;   // demand-allocated
      for(uint32 i=0; i<_tableSize; i++)
      {
         PortableHashtableEntry * next = &_table[i];
         if (next->_valid)
         {
            // Prepend static entry to our list of static entries
            next->_iterNext = staticHead;
            staticHead      = next;

            // And if there are any dynamic entries, prepend them to the dynamic list
            while((next = next->_next) != NULL)
            {
               next->_iterNext = dynamicHead;
               dynamicHead     = next;
            }
         }
      }
            
      // We now will move the static entries over first.
      // Since we may have to do dynamic allocations here, this part could 
      // fail.  So doing it first means there is less mess to clean up.
      while(staticHead != NULL)
      {
         PortableHashtableEntry * next  = staticHead->_iterNext;   // save this, before AddIterationEntry() munges it
         PortableHashtableEntry * newSlot = &newTable[staticHead->_hash % newTableSize];
         if (newSlot->_valid)
         {
            // Gotta convert our static entries into a dynamic one!
            PortableHashtableEntry * dyn = newnothrow PortableHashtableEntry(*staticHead);
            if (dyn)
            {
               dyn->_next = newSlot->_next;
               newSlot->_next = dyn;
               AddIterationEntry(dyn);
            }
            else
            {
               // oops -- clean up and go home
               WARN_OUT_OF_MEMORY;
               for (size_t i=0; i<newTableSize; i++)
               {
                  PortableHashtableEntry * n = &newTable[i];
                  if (n->_valid)
                  {
                     n = n->_next;
                     while(n)
                     {
                        PortableHashtableEntry * n2 = n->_next;
                        delete n;
                        n = n2;
                     }
                  }
               }
               delete [] newTable;
               RecalculateIterationEntries();  // since we munged these to make our lists
               return false;
            }
         }
         else
         {
            *newSlot = *staticHead;
            newSlot->_next = NULL;
            AddIterationEntry(newSlot);
         }
         staticHead = next;  // continue on...
      }

      // If we got here, we're home free.  Just move the dynamic entries over....
      while(dynamicHead != NULL)
      {
         PortableHashtableEntry * next    = dynamicHead->_iterNext;   // save this, before AddIterationEntry() munges it
         PortableHashtableEntry * newSlot = &newTable[dynamicHead->_hash % newTableSize];
         if (newSlot->_valid)
         {
            dynamicHead->_next = newSlot->_next;
            newSlot->_next = dynamicHead;
            AddIterationEntry(dynamicHead);
         }
         else
         {
            *newSlot = *dynamicHead;
            newSlot->_next = NULL;
            newSlot->_valid = true;
            AddIterationEntry(newSlot);
            delete dynamicHead;
         }
         dynamicHead = next;  // continue on...
      }

      // done!  Now switch over the tables
      delete [] _table;
      _table = newTable;
      _tableSize = newTableSize;
      _threshold = (uint32)(newTableSize * _loadFactor);
   }
   return true;  // success
}



// Redoes our linked list of entries for iterators
template <class KeyType, class ValueType>
void
PortableHashtable<KeyType,ValueType>::RecalculateIterationEntries()
{
   _iterHead = NULL;
   
   if (_table)
   {
      for(uint32 i = 0; i < _tableSize; i++)
      {
         PortableHashtableEntry * anchor = &_table[i];
         if (anchor->_valid)
         {
            PortableHashtableEntry * e = anchor;
            while(e) 
            {
               AddIterationEntry(e);
               e = e->_next;
            }
         }
      }
   }
}

// Add (e) to our iteration linked list
template <class KeyType, class ValueType>
void
PortableHashtable<KeyType,ValueType>::AddIterationEntry(PortableHashtableEntry * e)
{
   if (_iterHead)
   {
      e->_iterPrev = _iterHead;
      e->_iterNext = _iterHead->_iterNext;
      if (e->_iterNext) e->_iterNext->_iterPrev = e;
      _iterHead->_iterNext = e;
   }
   else
   {
      _iterHead = e;
      e->_iterPrev = e->_iterNext = NULL;
   }
}

// Remove (e) from our iteration linked list
template <class KeyType, class ValueType>
void 
PortableHashtable<KeyType,ValueType>::RemoveIterationEntry(PortableHashtableEntry * e)
{
   if (_iterHead == e) _iterHead = e->_iterNext;
   if (e->_iterPrev) e->_iterPrev->_iterNext = e->_iterNext;
   if (e->_iterNext) e->_iterNext->_iterPrev = e->_iterPrev;
   e->_iterPrev = e->_iterNext = NULL; 
}

template <class KeyType, class ValueType>
bool
PortableHashtable<KeyType,ValueType>::Put(const KeyType& key, const ValueType& value)
{
   return PutAux(key, value, NULL, NULL);  
}

template <class KeyType, class ValueType>
bool
PortableHashtable<KeyType,ValueType>::Put(const KeyType& key, const ValueType& value, ValueType & setPreviousValue, bool * optReplacedFlag)
{
   return PutAux(key, value, &setPreviousValue, optReplacedFlag);
}

template <class KeyType, class ValueType>
bool
PortableHashtable<KeyType,ValueType>::PutAux(const KeyType& key, const ValueType& value, ValueType * optSetPreviousValue, bool * optReplacedFlag)
{
   _stateCode+=100;

   if (optReplacedFlag) *optReplacedFlag = false;

   if (_table == NULL)  // demand-allocate this
   {
      _table = newnothrow PortableHashtableEntry[_tableSize];
      if (_table == NULL) {WARN_OUT_OF_MEMORY; return false;}
   }

   uint32 hash = (*_userHashFunc)(key);
   uint32 index = hash % _tableSize;

   // Make sure the key is not already in the hashtable
   PortableHashtableEntry * e = &_table[index];
   if (e->_valid)
   {
      while(e)
      {
         if ((e->_hash == hash)&&(KeysAreEqual(e->_key, key))) 
         {
            if (optSetPreviousValue) *optSetPreviousValue = e->_value;
            if (optReplacedFlag) *optReplacedFlag = true;
            e->_value = value;
            return true;
         }
         e = e->_next;
      }
   }

   // Rehash the table if the threshold is exceeded
   if (_count >= _threshold) return (GrowTable()) ? PutAux(key, value, optSetPreviousValue, optReplacedFlag) : false;

   PortableHashtableEntry * slot = &_table[index];
   if (slot->_valid)
   {
      // Creates new entry in linked list for this slot
      PortableHashtableEntry * newEntry = newnothrow PortableHashtableEntry();
      if (newEntry == NULL) {WARN_OUT_OF_MEMORY; return false;}
      newEntry->_hash  = hash;
      newEntry->_key   = key;
      newEntry->_value = value;
      newEntry->_valid = true;
      newEntry->_next  = slot->_next;
      slot->_next = newEntry;
      AddIterationEntry(newEntry);
   }
   else
   {
      // First entry in slot; just copy data over
      slot->_hash  = hash;
      slot->_key   = key;
      slot->_value = value;  
      slot->_next  = NULL;
      slot->_valid = true;
      AddIterationEntry(slot);
   }
   _count++;

   return true; // success!  (new semantics for MUSCLE v1.70)
}

template <class KeyType, class ValueType>
bool
PortableHashtable<KeyType,ValueType>::Remove(const KeyType& key)
{
   return RemoveAux(key, NULL);
}

template <class KeyType, class ValueType>
bool
PortableHashtable<KeyType,ValueType>::Remove(const KeyType& key, ValueType & setValue)
{
   return RemoveAux(key, &setValue);
}


template <class KeyType, class ValueType>
bool
PortableHashtable<KeyType,ValueType>::RemoveAux(const KeyType& key, ValueType * optSetValue)
{
   _stateCode+=10000;
   if (_table)
   {
      uint32 hash = (*_userHashFunc)(key);
      PortableHashtableEntry * e = &_table[hash % _tableSize];
      if (e->_valid)
      {
         PortableHashtableEntry * prev = NULL;
         while(e)
         {
            if ((e->_hash == hash)&&(KeysAreEqual(e->_key, key)))
            {
               if (optSetValue) *optSetValue = e->_value;
               PortableHashtableEntry * next = e->_next;
               if (prev)
               {
                  prev->_next = next;
                  RemoveIterationEntry(e);
                  delete e;
               }
               else
               {
                  if (next) 
                  {
                     RemoveIterationEntry(next);
                     *e = *next;
                     e->_valid = true;
                     delete next;
                  }
                  else 
                  {
                     RemoveIterationEntry(e);
                     e->Invalidate(_defaultKey, _defaultValue);
                  }
               }
               _count--;
               return true;
            }
            prev = e;
            e = e->_next;
         }
      }
   }
   return false;
}

template <class KeyType, class ValueType>
void
PortableHashtable<KeyType,ValueType>::Clear()
{
   _stateCode+=1000000;

   while(_iterHead)
   {
      PortableHashtableEntry * next = _iterHead->_iterNext;  // save for later
      if ((_iterHead >= _table)&&(_iterHead < _table+_tableSize)) _iterHead->Invalidate(_defaultKey, _defaultValue);
                                                             else delete _iterHead;
      _iterHead = next;
   }
   _count = 0;
}

template <class KeyType, class ValueType>
uint32
PortableHashtable<KeyType,ValueType>::NextPrime(uint32 start)
{
   if (start % 2 == 0) start++;
   uint32 i;
   for(; ; start += 2)
   {
      for(i = 3; i * i <= start; i += 2) if (start % i == 0) break;
      if (i * i > start) return start;
   }
}

//===============================================================
// Implementation of PortableHashtableIterator
//===============================================================

template <class KeyType, class ValueType>
PortableHashtableIterator<KeyType, ValueType>::PortableHashtableIterator() : _nextKeyCookie(NULL), _nextValueCookie(NULL), _stateCode(0), _owner(NULL)
{
   // empty
}

template <class KeyType, class ValueType>
PortableHashtableIterator<KeyType, ValueType>::PortableHashtableIterator(const PortableHashtableIterator<KeyType, ValueType> & rhs) 
{
   *this = rhs;
}

template <class KeyType, class ValueType>
PortableHashtableIterator<KeyType, ValueType>::PortableHashtableIterator(const PortableHashtable<KeyType, ValueType> & owner, uint32 stateCode) : _nextKeyCookie(owner._iterHead), _nextValueCookie(owner._iterHead), _stateCode(stateCode), _owner(&owner)
{
   // empty
}

template <class KeyType, class ValueType>
PortableHashtableIterator<KeyType, ValueType>::~PortableHashtableIterator()
{
   // empty
}

template <class KeyType, class ValueType>
bool
PortableHashtableIterator<KeyType,ValueType>::HasMoreKeys() const
{
   return (_nextKeyCookie != NULL);
}

template <class KeyType, class ValueType>
bool
PortableHashtableIterator<KeyType,ValueType>::HasMoreValues() const
{
   return (_nextValueCookie != NULL);
}

template <class KeyType, class ValueType>
bool
PortableHashtableIterator<KeyType,ValueType>::GetNextKey(KeyType& key) 
{
   KeyType * ret = GetNextKey();
   if (ret) 
   {
      key = *ret;
      return true;
   }
   else return false;
}

template <class KeyType, class ValueType>
bool
PortableHashtableIterator<KeyType,ValueType>::GetNextValue(ValueType& val)
{
   ValueType * ret = GetNextValue();
   if (ret)
   {
      val = *ret;
      return true;
   } 
   else return false;
}

template <class KeyType, class ValueType>
bool
PortableHashtableIterator<KeyType,ValueType>::PeekNextKey(KeyType& key) const
{
   KeyType * ret = PeekNextKey();
   if (ret)
   {
      key = *ret;
      return true;
   }
   else return false;
}

template <class KeyType, class ValueType>
bool
PortableHashtableIterator<KeyType,ValueType>::PeekNextValue(ValueType& val) const
{
   ValueType * ret = PeekNextValue();
   if (ret)
   {
      val = *ret;
      return true;
   }
   else return false;
}

template <class KeyType, class ValueType>
ValueType *
PortableHashtableIterator<KeyType,ValueType>::GetNextValue()
{
   ValueType * val = PeekNextValue();
   if (val)
   {
      _owner->IterateCookie(&_nextValueCookie);
      return val;
   }
   return NULL;
}

template <class KeyType, class ValueType>
KeyType *
PortableHashtableIterator<KeyType,ValueType>::GetNextKey() 
{
   KeyType * ret = PeekNextKey();
   if (ret)
   {
      _owner->IterateCookie(&_nextKeyCookie);
      return ret;
   }
   return NULL;
}

template <class KeyType, class ValueType>
ValueType *
PortableHashtableIterator<KeyType,ValueType>::PeekNextValue() const
{
   return (_owner) ? _owner->GetValueFromCookie(_nextValueCookie, _stateCode) : NULL;
}

template <class KeyType, class ValueType>
KeyType *
PortableHashtableIterator<KeyType,ValueType>::PeekNextKey() const
{
   return (_owner) ? _owner->GetKeyFromCookie(_nextKeyCookie, _stateCode) : NULL;
}

#endif

